#ifndef AST_HPP
#define AST_HPP

// expression

enum expressiontype
{
    EXP_FUNCALL,
    EXP_UNARY,
    EXP_BINARY,
    EXP_ASSIGN,
    EXP_COMPOUND,
    EXP_TERNARY,
    
    EXP_INT,
    EXP_CHAR,
    EXP_STRING,
    EXP_NAME
};

enum operatortype
{
    OP_NONE,
    
    // Unary
    OP_POSTINC,
    OP_POSTDEC,
    OP_PREINC,
    OP_PREDEC,
    OP_NEGATE,
    OP_NOT,
    OP_COMP,
    OP_DEREF,
    OP_ADDR,
    
    // Binary / compound
    OP_MODULO,
    OP_MULTIPLY,
    OP_DIVIDE,
    OP_SUBTRACT,
    OP_ADD,
    OP_LEFTSHIFT,
    OP_RIGHTSHIFT,
    OP_LESS,
    OP_LESSEQUAL,
    OP_GREATER,
    OP_GREATEREQUAL,
    OP_EQUAL,
    OP_NOTEQUAL,
    OP_AND,
    OP_XOR,
    OP_OR
};

class expressions;

class expression
{
    expression(expressiontype type);
    expression(const expression &); // = delete
    expression &operator =(const expression &); // = delete
    
public:
    expressiontype  type;
    int             ival;
    char           *sval;
    expression     *operands[3];
    expressions    *args;
    
    ~expression();
    
    static expression *make_funcall(expression *fun, expressions *args);
    static expression *make_unary(int operatortype, expression *operand);
    static expression *make_binary(int operatortype, expression *lhs, expression *rhs);
    static expression *make_assign(int operatortype, expression *lhs, expression *rhs);
    static expression *make_ternary(expression *lhs, expression *mid, expression *rhs);
    static expression *make_int(char *value);
    static expression *make_char(char *value);
    static expression *make_string(char *value);
    static expression *make_name(char *value);
    
    static expression *make_vector(expression *vec, expression *off);
};

// expressions

class expressions
{
    expressions(const expressions &); // = delete
    expressions &operator =(const expressions &); // = delete
    
public:
    expression  *expr;
    expressions *next;
    
    expressions(expression *expr, expressions *next) : expr(expr), next(next) {}
    ~expressions() { delete expr; delete next; }
};

// statement

enum statementtype
{
    STMT_AUTO,
    STMT_EXTRN,
    STMT_LABELLED,
    STMT_CASE,
    STMT_BLOCK,
    STMT_IF,
    STMT_WHILE,
    STMT_BREAK,
    STMT_SWITCH,
    STMT_GOTO,
    STMT_RETURN,
    STMT_NULL,
    STMT_EXPRESSION
};

class var_decl
{
    var_decl(const var_decl &); // = delete
    var_decl &operator =(const var_decl &); // = delete
    
public:
    char       *name;
    expression *vlen;
    var_decl   *next;
    
    var_decl(char *name, expression *init) : name(name), vlen(init), next(0) {}
    ~var_decl() { delete name; delete vlen; delete next; }
};

class statement
{
    statement(statementtype type);
    statement(const statement &); // = delete
    statement &operator =(const statement &); // = delete
    
public:
    statementtype  type;
    var_decl      *decls;
    char          *sval;
    expression    *expr;
    statement     *contained[2];
    statement     *next;
    
    ~statement();
    
    static statement *make_auto(var_decl *decls, statement *contained);
    static statement *make_extrn(var_decl *decls, statement *contained);
    static statement *make_labelled(char *label, statement *contained);
    static statement *make_case(expression *value, statement *contained);
    static statement *make_block(statement *contained);
    static statement *make_if(expression *condition, statement *then, statement *else_);
    static statement *make_while(expression *condition, statement *contained);
    static statement *make_break();
    static statement *make_switch(expression *value, statement *contained);
    static statement *make_goto(expression *value);
    static statement *make_return(expression *value);
    static statement *make_null();
    static statement *make_expression(expression *expr);
};

// definition

enum definitiontype
{
    DEF_SIMPLE,
    DEF_VECTOR,
    DEF_FUNCTION
};

class definition
{
    definition(definitiontype type, char *name);
    definition(const definition &); // = delete
    definition& operator =(const definition &); // = delete
    
public:
    definitiontype  type;
    char           *sval;
    expression     *vlen;
    expressions    *initialiser;
    var_decl       *args;
    statement      *stmt;
    definition     *next;
    
    ~definition();
    
    static definition *make_simple(char *name, expressions *initialiser);
    static definition *make_vector(char *name, expression *vlen, expressions *initialiser);
    static definition *make_function(char *name, var_decl *args, statement *stmt);
};

#endif
