#include "parser.hh"
#include "lexer.hh"

#include <cstdio>
#include <fstream>
#include <iostream>

bool compile(definition *program, std::ostream &fasm);

int yyparse(definition **program, yyscan_t yyscanner);

static const char *assign_sym(int op)
{
    switch (op)
    {
        case OP_OR: return "=|";
        case OP_AND: return "=&";
        case OP_EQUAL: return "===";
        case OP_NOTEQUAL: return "=!=";
        case OP_LESS: return "=<";
        case OP_LESSEQUAL: return "=<=";
        case OP_GREATER: return "=>";
        case OP_GREATEREQUAL: return "=>=";
        case OP_LEFTSHIFT: return "=<<";
        case OP_RIGHTSHIFT: return "=>>";
        case OP_SUBTRACT: return "=-";
        case OP_ADD: return "=+";
        case OP_MODULO: return "=%";
        case OP_MULTIPLY: return "=*";
        case OP_DIVIDE: return "=/";
        default: return "=";
    }
}

static void lex_dump(yyscan_t scanner, std::ostream &out)
{
    int tokentype;
    YYSTYPE tokenval;
    YYLTYPE location;
    
    while ((tokentype = yylex(&tokenval, &location, scanner)) > 0)
    {
        switch (tokentype)
        {
            case TK_OP_INCR: out << "Op ++\n"; break;
            case TK_OP_DECR: out << "Op --\n"; break;
            case TK_OP_NOT: out << "Op !\n"; break;
            
            case ';': out << "Op ;\n"; break;
            case '?': out << "Op ?\n"; break;
            case ':': out << "Op :\n"; break;
            case ',': out << "Op ,\n"; break;
            case '[': out << "Op [\n"; break;
            case ']': out << "Op ]\n"; break;
            case '(': out << "Op (\n"; break;
            case ')': out << "Op )\n"; break;
            case '{': out << "Op {\n"; break;
            case '}': out << "Op }\n"; break;
            
            case TK_OP_OR: out << "Op |\n"; break;
            case TK_OP_AND: out << "Op &\n"; break;
            case TK_OP_EQ: out << "Op ==\n"; break;
            case TK_OP_NEQ: out << "Op !=\n"; break;
            case TK_OP_LT: out << "Op <\n"; break;
            case TK_OP_LEQ: out << "Op <=\n"; break;
            case TK_OP_GT: out << "Op >\n"; break;
            case TK_OP_GEQ: out << "Op >=\n"; break;
            case TK_OP_LS: out << "Op <<\n"; break;
            case TK_OP_RS: out << "Op >>\n"; break;
            case TK_OP_SUB: out << "Op -\n"; break;
            case TK_OP_ADD: out << "Op +\n"; break;
            case TK_OP_MOD: out << "Op %%\n"; break;
            case TK_OP_MULT: out << "Op *\n"; break;
            case TK_OP_DIV: out << "Op /\n"; break;
            
            case TK_ASSIGNMENT:
                out << "As " << assign_sym(tokenval.ival) << "\n";
                break;
                
            case TK_KW_AUTO: out << "Kw auto\n"; break;
            case TK_KW_CASE: out << "Kw case\n"; break;
            case TK_KW_ELSE: out << "Kw else\n"; break;
            case TK_KW_EXTRN: out << "Kw extrn\n"; break;
            case TK_KW_GOTO: out << "Kw goto\n"; break;
            case TK_KW_IF: out << "Kw if\n"; break;
            case TK_KW_RETURN: out << "Kw return\n"; break;
            case TK_KW_SWITCH: out << "Kw switch\n"; break;
            case TK_KW_WHILE: out << "Kw while\n"; break;
                
            case TK_NAME:
                out << "Id " << tokenval.sval << "\n";
                delete tokenval.sval;
                break;
                
            case TK_INT:
            case TK_CHAR:
            case TK_STRING:
                out << "Ct " << tokenval.sval << "\n";
                delete tokenval.sval;
                break;
        }
    }
}

static void dump_statement(statement *stmt, std::ostream &out, int indent)
{
    indent++;
    
    for ( ; stmt; stmt = stmt->next)
    {
        for (int i = 0; i < indent; i++)
            out << "  ";
        
        switch (stmt->type)
        {
            case STMT_AUTO:
                out << "auto\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_EXTRN:
                out << "extrn\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_LABELLED:
                out << "label\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_CASE:
                out << "case\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_BLOCK:
                out << "block\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_IF:
                out << "if\n";
                dump_statement(stmt->contained[0], out, indent);
                if (stmt->contained[1])
                {
                    for (int i = 0; i < indent; i++)
                        out << "  ";
                    out << "else\n";
                    dump_statement(stmt->contained[1], out, indent);
                }
                break;
            case STMT_WHILE:
                out << "while\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_SWITCH:
                out << "switch\n";
                dump_statement(stmt->contained[0], out, indent);
                break;
            case STMT_GOTO:
                out << "goto\n";
                break;
            case STMT_RETURN:
                out << "return\n";
                break;
            case STMT_NULL:
                out << "null\n";
                break;
            case STMT_EXPRESSION:
                out << "expression\n";
                break;
            default:
                out << "ERROR!\n";
                break;
        }
    }
}

static void dump_program(definition *program, std::ostream &out, int detail)
{
    definition *def;
    
    for (def = program; def; def = def->next)
    {
        out << def->sval << " ";
        switch (def->type)
        {
            case DEF_SIMPLE:
                out << "simple\n";
                break;
            case DEF_VECTOR:
                out << "vector\n";
                break;
            case DEF_FUNCTION:
                out << "function\n";
                if (detail)
                    dump_statement(def->stmt, out, 0);
                break;
        }
    }
}

int main(int argc, char **argv)
{
    const char *inname = NULL, *outname = NULL;
    enum { COMPILE, LEX, DEFS, STMTS } mode = COMPILE;
    
    int i;
    
    for (i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "-o") == 0)
        {
            if (++i == argc)
            {
                std::cerr << "Option -o requires an argument" << std::endl;
                return 2;
            }
            outname = argv[i];
        }
        else if (strcmp(argv[i], "-d") == 0)
        {
            if (++i == argc)
            {
                std::cerr << "Option -d requires an argument" << std::endl;
                return 2;
            }
            else if (strcmp(argv[i], "lex") == 0)
            {
                mode = LEX;
            }
            else if (strcmp(argv[i], "defs") == 0)
            {
                mode = DEFS;
            }
            else if (strcmp(argv[i], "stmts") == 0)
            {
                mode = STMTS;
            }
            else
            {
                std::cerr << "Invalid argument to -d" << std::endl;
                return 2;
            }
        }
        else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
        {
            std::cout << argv[0] << " [-o outfile] [-d lex|defs|stmts] [infile]\n"
                      << argv[0] << " -h\n"
                      << "\n"
                      << "  Reads b source code from infile (or standard input), parses and compiles it to assembly which is written to outfile (or standard output). With -h, instead output this help.\n"
                      << "  With -d lex, only perform lexical analysis on the input and output the list of tokens.\n"
                      << "  With -d defs or -d stmts, parse the source code and output the list of external definitions. -d stmts also writes the tree of statements within each function." << std::endl;
            return 0;
        }
        else
        {
            inname = argv[i];
        }
    }
    
    FILE *infile = NULL;
    if (inname)
    {
        infile = fopen(inname, "r");
        if (!infile)
        {
            std::cerr << "Could not open '" << inname << "' for reading" << std::endl;
            return 1;
        }
    }
    
    std::ofstream named_outfile;
    if (outname)
    {
        named_outfile.open(outname);
        if (!named_outfile)
        {
            std::cerr << "Could not open '" << outname << "' for reading" << std::endl;
            return 1;
        }
    }
    std::ostream &outfile = outname ? named_outfile : std::cout;
    
    yyscan_t scanner;
    yylex_init(&scanner);
    if (infile)
        yyset_in(infile, scanner);
    
    definition *program = NULL;
    if (mode == LEX)
        lex_dump(scanner, outfile);
    else
        if (yyparse(&program, scanner) != 0)
            return 1;
    
    yylex_destroy(scanner);
    
    if (mode == DEFS)
        dump_program(program, outfile, 0);
    else if (mode == STMTS)
        dump_program(program, outfile, 1);
    else if (mode == COMPILE)
        if (!compile(program, outfile))
            return 1;
    
    delete program;
    
    if (inname)
        fclose(infile);
    
    return 0;
}
