%code requires {
#include "ast.hh"
}

%output "parser.cc"
%defines "parser.hh"

%`define' api.pure
%locations

%lex-param { yyscan_t yyscanner }
%parse-param { definition **program }
%parse-param { yyscan_t yyscanner }

%union {
    int          ival;
    char        *sval;
    expression  *expr;
    expressions *exprs;
    var_decl    *decl;
    statement   *stmt;
    definition  *defn;
}

%{
#include "lexer.hh"

#include <iostream>

void yyerror(YYLTYPE *locp, definition ** /*program*/, yyscan_t /*yyscanner*/, const char *msg)
{
    std::cerr << msg << " on line " << locp->first_line << std::endl;
}
%}

%token TK_OP_INCR
%token TK_OP_DECR
%token TK_OP_NOT
ifdef(`EXTENSIONS', %token TK_OP_COMP)

%token TK_OP_OR
ifdef(`EXTENSIONS', %token TK_OP_XOR)
%token TK_OP_AND
%token TK_OP_EQ
%token TK_OP_NEQ
%token TK_OP_LT
%token TK_OP_LEQ
%token TK_OP_GT
%token TK_OP_GEQ
%token TK_OP_LS
%token TK_OP_RS
%token TK_OP_SUB
%token TK_OP_ADD
%token TK_OP_MOD
%token TK_OP_MULT
%token TK_OP_DIV

%token <ival> TK_ASSIGNMENT

%token TK_KW_AUTO
ifdef(`EXTENSIONS', %token TK_KW_BREAK)
%token TK_KW_CASE
%token TK_KW_ELSE
%token TK_KW_EXTRN
%token TK_KW_GOTO
%token TK_KW_IF
%token TK_KW_RETURN
%token TK_KW_SWITCH
%token TK_KW_WHILE

%token <sval> TK_NAME

%token <sval> TK_INT
%token <sval> TK_CHAR
%token <sval> TK_STRING

%type <defn> program
%type <defn> definition
%type <exprs> ivals_opt
%type <exprs> ivals
%type <decl> names
%type <expr> ival
%type <stmt> statement
%type <stmt> statement_notvalue
%type <decl> vardecl
%type <decl> vardecls
%type <stmt> statements_opt
%type <expr> value
%type <exprs> values
%type <expr> constant
%type <expr> name

%nonassoc TK_KW_IF
%nonassoc TK_KW_ELSE

%right TK_ASSIGNMENT
%right '?' ':'
%left TK_OP_OR
ifdef(`EXTENSIONS', %left TK_OP_XOR)
%left TK_OP_AND
%left TK_OP_EQ TK_OP_NEQ
%left TK_OP_LT TK_OP_LEQ TK_OP_GT TK_OP_GEQ
%left TK_OP_LS TK_OP_RS
%left TK_OP_SUB TK_OP_ADD
%left TK_OP_MOD TK_OP_MULT TK_OP_DIV
%right PREFIX
%right SUFFIX TK_OP_INCR TK_OP_DECR '(' '['

%%

program:                    { *program = $$ = 0; }
       | definition program { *program = $$ = $1; $$->next = $2; }
       ;

definition: TK_NAME ivals_opt ';'                  { $$ = definition::make_simple($1, $2); }
          | TK_NAME '[' ']' ivals_opt ';'          { $$ = definition::make_vector($1, 0, $4); }
          | TK_NAME '[' constant ']' ivals_opt ';' { $$ = definition::make_vector($1, $3, $5); }
          | TK_NAME '(' ')' statement              { $$ = definition::make_function($1, 0, $4); }
          | TK_NAME '(' names ')' statement        { $$ = definition::make_function($1, $3, $5); }
          ;

ivals_opt:       { $$ = 0; }
         | ivals
         ;

ivals: ival           { $$ = new expressions($1, 0); }
     | ival ',' ivals { $$ = new expressions($1, $3); }
     ;

names: TK_NAME           { $$ = new var_decl($1, 0); }
     | TK_NAME ',' names { $$ = new var_decl($1, 0); $$->next = $3; }
     ;

ival: constant
    | name
    ;

statement_notvalue: TK_KW_AUTO vardecls ';' statement                     { $$ = statement::make_auto($2, $4); }
                  | TK_KW_EXTRN names ';' statement                       { $$ = statement::make_extrn($2, $4); }
                  | TK_NAME ':' statement                                 { $$ = statement::make_labelled($1, $3); }
                  | TK_KW_CASE constant ':' statement                     { $$ = statement::make_case($2, $4); }
                  | '{' statements_opt '}'                                { $$ = statement::make_block($2); }
                  | TK_KW_IF '(' value ')' statement %prec TK_KW_IF       { $$ = statement::make_if($3, $5, 0); }
                  | TK_KW_IF '(' value ')' statement TK_KW_ELSE statement { $$ = statement::make_if($3, $5, $7); }
                  | TK_KW_WHILE '(' value ')' statement                   { $$ = statement::make_while($3, $5); }
ifdef(`EXTENSIONS',
`                  | TK_KW_BREAK                                           { $$ = statement::make_break(); }')
                  | TK_KW_SWITCH value statement_notvalue                 { $$ = statement::make_switch($2, $3); }
                  | TK_KW_GOTO value ';'                                  { $$ = statement::make_goto($2); }
                  | TK_KW_RETURN ';'                                      { $$ = statement::make_return(0); }
                  | TK_KW_RETURN '(' value ')' ';'                        { $$ = statement::make_return($3); }
                  | ';'                                                   { $$ = statement::make_null(); }
                  ;

statement: statement_notvalue
         | value ';'          { $$ = statement::make_expression($1); }
         ;

vardecl: TK_NAME                  { $$ = new var_decl($1, 0); }
ifdef(`EXTENSIONS',
`       '| TK_NAME '[' constant ']' { $$ = new var_decl($1, $3); },
`       | TK_NAME constant         { $$ = new var_decl($1, $2); }')
       ;

vardecls: vardecl
        | vardecl ',' vardecls { $$ = $1; $$->next = $3; }
        ;

statements_opt:                          { $$ = 0; }
              | statement statements_opt { $$ = $1; $$->next = $2; }
              ;

value: '(' value ')'                     { $$ = $2; }
     | constant
     | value TK_ASSIGNMENT value         { $$ = expression::make_assign($2, $1, $3); }
     | TK_OP_INCR value %prec PREFIX     { $$ = expression::make_unary(OP_PREINC, $2); }
     | TK_OP_DECR value %prec PREFIX     { $$ = expression::make_unary(OP_PREDEC, $2); }
     | value TK_OP_INCR %prec SUFFIX     { $$ = expression::make_unary(OP_POSTINC, $1); }
     | value TK_OP_DECR %prec SUFFIX     { $$ = expression::make_unary(OP_POSTDEC, $1); }
     | TK_OP_SUB value %prec PREFIX      { $$ = expression::make_unary(OP_NEGATE, $2); }
     | TK_OP_NOT value %prec PREFIX      { $$ = expression::make_unary(OP_NOT, $2); }
ifdef(`EXTENSIONS',
`     | TK_OP_COMP value %prec PREFIX     { $$ = expression::make_unary(OP_COMP, $2); }')
     | TK_OP_AND value %prec PREFIX      { $$ = expression::make_unary(OP_ADDR, $2); }
     | value TK_OP_OR value              { $$ = expression::make_binary(OP_OR, $1, $3); }
ifdef(`EXTENSIONS',
`     | value TK_OP_XOR value             { $$ = expression::make_binary(OP_XOR, $1, $3); }')
     | value TK_OP_AND value             { $$ = expression::make_binary(OP_AND, $1, $3); }
     | value TK_OP_EQ value              { $$ = expression::make_binary(OP_EQUAL, $1, $3); }
     | value TK_OP_NEQ value             { $$ = expression::make_binary(OP_NOTEQUAL, $1, $3); }
     | value TK_OP_LT value              { $$ = expression::make_binary(OP_LESS, $1, $3); }
     | value TK_OP_LEQ value             { $$ = expression::make_binary(OP_LESSEQUAL, $1, $3); }
     | value TK_OP_GT value              { $$ = expression::make_binary(OP_GREATER, $1, $3); }
     | value TK_OP_GEQ value             { $$ = expression::make_binary(OP_GREATEREQUAL, $1, $3); }
     | value TK_OP_LS value              { $$ = expression::make_binary(OP_LEFTSHIFT, $1, $3); }
     | value TK_OP_RS value              { $$ = expression::make_binary(OP_RIGHTSHIFT, $1, $3); }
     | value TK_OP_SUB value             { $$ = expression::make_binary(OP_SUBTRACT, $1, $3); }
     | value TK_OP_ADD value             { $$ = expression::make_binary(OP_ADD, $1, $3); }
     | value TK_OP_MOD value             { $$ = expression::make_binary(OP_MODULO, $1, $3); }
     | value TK_OP_MULT value            { $$ = expression::make_binary(OP_MULTIPLY, $1, $3); }
     | value TK_OP_DIV value             { $$ = expression::make_binary(OP_DIVIDE, $1, $3); }
     | value '?' value ':' value         { $$ = expression::make_ternary($1, $3, $5); }
     | value '(' ')' %prec SUFFIX        { $$ = expression::make_funcall($1, 0); }
     | value '(' values ')' %prec SUFFIX { $$ = expression::make_funcall($1, $3); }
     | name
     | TK_OP_MULT value %prec PREFIX     { $$ = expression::make_unary(OP_DEREF, $2); }
     | value '[' value ']' %prec SUFFIX  { $$ = expression::make_vector($1, $3); }
     ;

values: value            { $$ = new expressions($1, 0); }
      | value ',' values { $$ = new expressions($1, $3); }
      ;

constant: TK_INT    { $$ = expression::make_int($1); }
        | TK_CHAR   { $$ = expression::make_char($1); }
        | TK_STRING { $$ = expression::make_string($1); }
        ;

name: TK_NAME { $$ = expression::make_name($1); }
    ;
